"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator.throw(value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments)).next());
    });
};
const MysqlSchemaBuilder_1 = require("../schema-builder/MysqlSchemaBuilder");
const ConnectionIsNotSetError_1 = require("./error/ConnectionIsNotSetError");
const BaseDriver_1 = require("./BaseDriver");
/**
 * This driver organizes work with mysql database.
 */
class MysqlDriver extends BaseDriver_1.BaseDriver {
    // -------------------------------------------------------------------------
    // Constructor
    // -------------------------------------------------------------------------
    constructor(connectionOptions, mysql) {
        super();
        this.connectionOptions = connectionOptions;
        /**
         * Pool of database connections.
         */
        this.databaseConnectionPool = [];
        // if driver dependency is not given explicitly, then try to load it via "require"
        if (!mysql && require) {
            try {
                mysql = require("mysql");
            }
            catch (e) {
                throw new Error("Mysql package has not been found installed. Try to install it: npm install mysql --save");
            }
        }
        else {
            throw new Error("Cannot load mysql driver dependencies. Try to install all required dependencies.");
        }
        this.mysql = mysql;
    }
    // -------------------------------------------------------------------------
    // Getter Methods
    // -------------------------------------------------------------------------
    escapeColumnName(columnName) {
        return columnName;
    }
    escapeAliasName(aliasName) {
        return aliasName;
        // return "\"" + aliasName + "\"";
    }
    escapeTableName(tableName) {
        return tableName;
        // return "\"" + aliasName + "\"";
    }
    retrieveDatabaseConnection() {
        if (this.pool) {
            return new Promise((ok, fail) => {
                this.pool.getConnection((err, connection) => {
                    if (err) {
                        fail(err);
                        return;
                    }
                    let dbConnection = this.databaseConnectionPool.find(dbConnection => dbConnection.connection === connection);
                    if (!dbConnection) {
                        dbConnection = {
                            id: this.databaseConnectionPool.length,
                            connection: connection,
                            isTransactionActive: false
                        };
                        this.databaseConnectionPool.push(dbConnection);
                    }
                    ok(dbConnection);
                });
            });
        }
        if (this.databaseConnection)
            return Promise.resolve(this.databaseConnection);
        throw new ConnectionIsNotSetError_1.ConnectionIsNotSetError("mysql");
    }
    releaseDatabaseConnection(dbConnection) {
        if (this.pool)
            dbConnection.connection.release();
        return Promise.resolve();
    }
    /**
     * Access to the native implementation of the database.
     */
    get native() {
        return {
            driver: this.mysql,
            connection: this.databaseConnection ? this.databaseConnection.connection : undefined,
            pool: this.pool
        };
    }
    /**
     * Access to the native connection to the database.
     */
    get nativeConnection() {
        return this.databaseConnection;
    }
    /**
     * Database name to which this connection is made.
     */
    get db() {
        if (this.databaseConnection && this.databaseConnection.connection.config.database)
            return this.databaseConnection.connection.config.database;
        if (this.connectionOptions.database)
            return this.connectionOptions.database;
        throw new Error("Cannot get the database name. Since database name is not explicitly given in configuration " +
            "(maybe connection url is used?), database name cannot be retrieved until connection is made.");
    }
    // -------------------------------------------------------------------------
    // Public Methods
    // -------------------------------------------------------------------------
    /**
     * Creates a schema builder which can be used to build database/table schemas.
     */
    createSchemaBuilder(dbConnection) {
        return new MysqlSchemaBuilder_1.MysqlSchemaBuilder(this, dbConnection);
    }
    /**
     * Performs connection to the database based on given connection options.
     */
    connect() {
        const options = Object.assign({}, {
            host: this.connectionOptions.host,
            user: this.connectionOptions.username,
            password: this.connectionOptions.password,
            database: this.connectionOptions.database,
            port: this.connectionOptions.port
        }, this.connectionOptions.extra || {});
        if (this.connectionOptions.usePool === false) {
            return new Promise((ok, fail) => {
                this.databaseConnection = {
                    id: 1,
                    connection: this.mysql.createConnection(options),
                    isTransactionActive: false
                };
                this.databaseConnection.connection.connect((err) => err ? fail(err) : ok());
            });
        }
        else {
            this.pool = this.mysql.createPool(options);
            return Promise.resolve();
        }
    }
    /**
     * Closes connection with database.
     */
    disconnect() {
        this.checkIfConnectionSet();
        return new Promise((ok, fail) => {
            const handler = (err) => err ? fail(err) : ok();
            if (this.pool) {
                this.pool.end(handler);
                this.pool = undefined;
                this.databaseConnectionPool = [];
            }
            if (this.databaseConnection) {
                this.databaseConnection.connection.end(handler);
                this.databaseConnection = undefined;
            }
        });
    }
    /**
     * Escapes given value.
     */
    escape(dbConnection, value) {
        return dbConnection.connection.escape(value);
    }
    /**
     * Executes a given SQL query.
     */
    query(dbConnection, query, parameters) {
        this.checkIfConnectionSet();
        this.logQuery(query);
        return new Promise((ok, fail) => dbConnection.connection.query(query, parameters, (err, result) => {
            if (err) {
                this.logFailedQuery(query);
                this.logQueryError(err);
                fail(err);
            }
            else {
                // console.log(`OK [${dbConnection.id}]: `, query);
                ok(result);
            }
        }));
    }
    /**
     * Starts transaction.
     */
    beginTransaction(dbConnection) {
        return __awaiter(this, void 0, void 0, function* () {
            if (dbConnection.isTransactionActive)
                throw new Error(`Transaction already started for the given connection, commit current transaction before starting a new one.`);
            yield this.query(dbConnection, "START TRANSACTION");
            dbConnection.isTransactionActive = true;
        });
    }
    /**
     * Commits transaction.
     */
    commitTransaction(dbConnection) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!dbConnection.isTransactionActive)
                throw new Error(`Transaction is not started yet, start transaction before committing it.`);
            yield this.query(dbConnection, "COMMIT");
            dbConnection.isTransactionActive = false;
        });
    }
    /**
     * Rollbacks transaction.
     */
    rollbackTransaction(dbConnection) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!dbConnection.isTransactionActive)
                throw new Error(`Transaction is not started yet, start transaction before rolling it back.`);
            yield this.query(dbConnection, "ROLLBACK");
            dbConnection.isTransactionActive = false;
        });
    }
    /**
     * Clears all tables in the currently connected database.
     */
    clearDatabase(dbConnection) {
        return __awaiter(this, void 0, void 0, function* () {
            this.checkIfConnectionSet();
            const disableForeignKeysCheckQuery = `SET FOREIGN_KEY_CHECKS = 0;`;
            const dropTablesQuery = `SELECT concat('DROP TABLE IF EXISTS ', table_name, ';') AS q FROM ` +
                `information_schema.tables WHERE table_schema = '${this.db}';`;
            const enableForeignKeysCheckQuery = `SET FOREIGN_KEY_CHECKS = 1;`;
            return this.query(dbConnection, disableForeignKeysCheckQuery)
                .then(() => this.query(dbConnection, dropTablesQuery))
                .then(results => Promise.all(results.map(q => this.query(dbConnection, q["q"]))))
                .then(() => this.query(dbConnection, enableForeignKeysCheckQuery))
                .then(() => { });
        });
    }
    buildParameters(sql, parameters) {
        if (!parameters || !Object.keys(parameters).length)
            return [];
        const builtParameters = [];
        const keys = Object.keys(parameters).map(parameter => "(:" + parameter + ")").join("|");
        sql.replace(new RegExp(keys, "g"), (key) => {
            const value = parameters[key.substr(1)];
            builtParameters.push(value);
            return "?";
        }); // todo: make replace only in value statements, otherwise problems
        return builtParameters;
    }
    replaceParameters(sql, parameters) {
        if (!parameters || !Object.keys(parameters).length)
            return sql;
        const keys = Object.keys(parameters).map(parameter => "(:" + parameter + ")").join("|");
        return sql.replace(new RegExp(keys, "g"), "?");
    }
    // -------------------------------------------------------------------------
    // Protected Methods
    // -------------------------------------------------------------------------
    checkIfConnectionSet() {
        if (!this.databaseConnection && !this.pool)
            throw new ConnectionIsNotSetError_1.ConnectionIsNotSetError("mysql");
    }
}
exports.MysqlDriver = MysqlDriver;

//# sourceMappingURL=MysqlDriver.js.map
